import pickle

from ...types.enum import Enum


class PickleEnum(Enum):
    # is defined outside of test because pickle unable to dump class inside ot pytest function
    A = "a"
    B = 1


def test_enums_pickling():
    a = PickleEnum.A
    pickled = pickle.dumps(a)
    restored = pickle.loads(pickled)
    assert type(a) is type(restored)
    assert a == restored
    assert a.value == restored.value
    assert a.name == restored.name

    b = PickleEnum.B
    pickled = pickle.dumps(b)
    restored = pickle.loads(pickled)
    assert type(a) is type(restored)
    assert b == restored
    assert b.value == restored.value
    assert b.name == restored.name



from graphene.test import Client

from ..data import setup
from ..schema import schema

setup()

client = Client(schema)


def test_mutations():
    result = client.execute("""
        mutation MyMutation {
          introduceShip(input:{clientMutationId:"abc", shipName: "Peter", factionId: "1"}) {
            ship {
              id
              name
            }
            faction {
              name
              ships {
                edges {
                  node {
                    id
                    name
                  }
                }
              }
            }
          }
        }
    """)
    assert result == {
        "data": {
            "introduceShip": {
                "ship": {"id": "U2hpcDo5", "name": "Peter"},
                "faction": {
                    "name": "Alliance to Restore the Republic",
                    "ships": {
                        "edges": [
                            {"node": {"id": "U2hpcDox", "name": "X-Wing"}},
                            {"node": {"id": "U2hpcDoy", "name": "Y-Wing"}},
                            {"node": {"id": "U2hpcDoz", "name": "A-Wing"}},
                            {"node": {"id": "U2hpcDo0", "name": "Millennium Falcon"}},
                            {"node": {"id": "U2hpcDo1", "name": "Home One"}},
                            {"node": {"id": "U2hpcDo5", "name": "Peter"}},
                        ]
                    },
                },
            }
        }
    }



human_data = {}
droid_data = {}


def setup():
    from .schema import Human, Droid

    global human_data, droid_data
    luke = Human(
        id="1000",
        name="Luke Skywalker",
        friends=["1002", "1003", "2000", "2001"],
        appears_in=[4, 5, 6],
        home_planet="Tatooine",
    )

    vader = Human(
        id="1001",
        name="Darth Vader",
        friends=["1004"],
        appears_in=[4, 5, 6],
        home_planet="Tatooine",
    )

    han = Human(
        id="1002",
        name="Han Solo",
        friends=["1000", "1003", "2001"],
        appears_in=[4, 5, 6],
        home_planet=None,
    )

    leia = Human(
        id="1003",
        name="Leia Organa",
        friends=["1000", "1002", "2000", "2001"],
        appears_in=[4, 5, 6],
        home_planet="Alderaan",
    )

    tarkin = Human(
        id="1004",
        name="Wilhuff Tarkin",
        friends=["1001"],
        appears_in=[4],
        home_planet=None,
    )

    human_data = {
        "1000": luke,
        "1001": vader,
        "1002": han,
        "1003": leia,
        "1004": tarkin,
    }

    c3po = Droid(
        id="2000",
        name="C-3PO",
        friends=["1000", "1002", "1003", "2001"],
        appears_in=[4, 5, 6],
        primary_function="Protocol",
    )

    r2d2 = Droid(
        id="2001",
        name="R2-D2",
        friends=["1000", "1002", "1003"],
        appears_in=[4, 5, 6],
        primary_function="Astromech",
    )

    droid_data = {"2000": c3po, "2001": r2d2}


def get_character(id):
    return human_data.get(id) or droid_data.get(id)


def get_friends(character):
    return map(get_character, character.friends)


def get_hero(episode):
    if episode == 5:
        return human_data["1000"]
    return droid_data["2001"]


def get_human(id):
    return human_data.get(id)


def get_droid(id):
    return droid_data.get(id)



from graphene.test import Client

from ..data import setup
from ..schema import schema

setup()

client = Client(schema)


def test_hero_name_query():
    result = client.execute("""
        query HeroNameQuery {
          hero {
            name
          }
        }
    """)
    assert result == {"data": {"hero": {"name": "R2-D2"}}}


def test_hero_name_and_friends_query():
    result = client.execute("""
        query HeroNameAndFriendsQuery {
          hero {
            id
            name
            friends {
              name
            }
          }
        }
    """)
    assert result == {
        "data": {
            "hero": {
                "id": "2001",
                "name": "R2-D2",
                "friends": [
                    {"name": "Luke Skywalker"},
                    {"name": "Han Solo"},
                    {"name": "Leia Organa"},
                ],
            }
        }
    }


def test_nested_query():
    result = client.execute("""
        query NestedQuery {
          hero {
            name
            friends {
              name
              appearsIn
              friends {
                name
              }
            }
          }
        }
    """)
    assert result == {
        "data": {
            "hero": {
                "name": "R2-D2",
                "friends": [
                    {
                        "name": "Luke Skywalker",
                        "appearsIn": ["NEWHOPE", "EMPIRE", "JEDI"],
                        "friends": [
                            {"name": "Han Solo"},
                            {"name": "Leia Organa"},
                            {"name": "C-3PO"},
                            {"name": "R2-D2"},
                        ],
                    },
                    {
                        "name": "Han Solo",
                        "appearsIn": ["NEWHOPE", "EMPIRE", "JEDI"],
                        "friends": [
                            {"name": "Luke Skywalker"},
                            {"name": "Leia Organa"},
                            {"name": "R2-D2"},
                        ],
                    },
                    {
                        "name": "Leia Organa",
                        "appearsIn": ["NEWHOPE", "EMPIRE", "JEDI"],
                        "friends": [
                            {"name": "Luke Skywalker"},
                            {"name": "Han Solo"},
                            {"name": "C-3PO"},
                            {"name": "R2-D2"},
                        ],
                    },
                ],
            }
        }
    }


def test_fetch_luke_query():
    result = client.execute("""
        query FetchLukeQuery {
          human(id: "1000") {
            name
          }
        }
    """)
    assert result == {"data": {"human": {"name": "Luke Skywalker"}}}


def test_fetch_some_id_query():
    result = client.execute(
        """
        query FetchSomeIDQuery($someId: String!) {
          human(id: $someId) {
            name
          }
        }
    """,
        variables={"someId": "1000"},
    )
    assert result == {"data": {"human": {"name": "Luke Skywalker"}}}


def test_fetch_some_id_query2():
    result = client.execute(
        """
        query FetchSomeIDQuery($someId: String!) {
          human(id: $someId) {
            name
          }
        }
    """,
        variables={"someId": "1002"},
    )
    assert result == {"data": {"human": {"name": "Han Solo"}}}


def test_invalid_id_query():
    result = client.execute(
        """
        query humanQuery($id: String!) {
          human(id: $id) {
            name
          }
        }
    """,
        variables={"id": "not a valid id"},
    )
    assert result == {"data": {"human": None}}


def test_fetch_luke_aliased():
    result = client.execute("""
        query FetchLukeAliased {
          luke: human(id: "1000") {
            name
          }
        }
    """)
    assert result == {"data": {"luke": {"name": "Luke Skywalker"}}}


def test_fetch_luke_and_leia_aliased():
    result = client.execute("""
        query FetchLukeAndLeiaAliased {
          luke: human(id: "1000") {
            name
          }
          leia: human(id: "1003") {
            name
          }
        }
    """)
    assert result == {
        "data": {"luke": {"name": "Luke Skywalker"}, "leia": {"name": "Leia Organa"}}
    }


def test_duplicate_fields():
    result = client.execute("""
        query DuplicateFields {
          luke: human(id: "1000") {
            name
            homePlanet
          }
          leia: human(id: "1003") {
            name
            homePlanet
          }
        }
    """)
    assert result == {
        "data": {
            "luke": {"name": "Luke Skywalker", "homePlanet": "Tatooine"},
            "leia": {"name": "Leia Organa", "homePlanet": "Alderaan"},
        }
    }


def test_use_fragment():
    result = client.execute("""
        query UseFragment {
          luke: human(id: "1000") {
            ...HumanFragment
          }
          leia: human(id: "1003") {
            ...HumanFragment
          }
        }
        fragment HumanFragment on Human {
          name
          homePlanet
        }
    """)
    assert result == {
        "data": {
            "luke": {"name": "Luke Skywalker", "homePlanet": "Tatooine"},
            "leia": {"name": "Leia Organa", "homePlanet": "Alderaan"},
        }
    }


def test_check_type_of_r2():
    result = client.execute("""
        query CheckTypeOfR2 {
          hero {
            __typename
            name
          }
        }
    """)
    assert result == {"data": {"hero": {"__typename": "Droid", "name": "R2-D2"}}}


def test_check_type_of_luke():
    result = client.execute("""
        query CheckTypeOfLuke {
          hero(episode: EMPIRE) {
            __typename
            name
          }
        }
    """)
    assert result == {
        "data": {"hero": {"__typename": "Human", "name": "Luke Skywalker"}}
    }



import graphene
from graphene import relay

from .data import create_ship, get_empire, get_faction, get_rebels, get_ship


class Ship(graphene.ObjectType):
    """A ship in the Star Wars saga"""

    class Meta:
        interfaces = (relay.Node,)

    name = graphene.String(description="The name of the ship.")

    @classmethod
    def get_node(cls, info, id):
        return get_ship(id)


class ShipConnection(relay.Connection):
    class Meta:
        node = Ship


class Faction(graphene.ObjectType):
    """A faction in the Star Wars saga"""

    class Meta:
        interfaces = (relay.Node,)

    name = graphene.String(description="The name of the faction.")
    ships = relay.ConnectionField(
        ShipConnection, description="The ships used by the faction."
    )

    def resolve_ships(self, info, **args):
        # Transform the instance ship_ids into real instances
        return [get_ship(ship_id) for ship_id in self.ships]

    @classmethod
    def get_node(cls, info, id):
        return get_faction(id)


class IntroduceShip(relay.ClientIDMutation):
    class Input:
        ship_name = graphene.String(required=True)
        faction_id = graphene.String(required=True)

    ship = graphene.Field(Ship)
    faction = graphene.Field(Faction)

    @classmethod
    def mutate_and_get_payload(
        cls, root, info, ship_name, faction_id, client_mutation_id=None
    ):
        ship = create_ship(ship_name, faction_id)
        faction = get_faction(faction_id)
        return IntroduceShip(ship=ship, faction=faction)


class Query(graphene.ObjectType):
    rebels = graphene.Field(Faction)
    empire = graphene.Field(Faction)
    node = relay.Node.Field()

    def resolve_rebels(root, info):
        return get_rebels()

    def resolve_empire(root, info):
        return get_empire()


class Mutation(graphene.ObjectType):
    introduce_ship = IntroduceShip.Field()


schema = graphene.Schema(query=Query, mutation=Mutation)



# coding: utf-8
from ..str_converters import to_camel_case, to_snake_case


def test_snake_case():
    assert to_snake_case("snakesOnAPlane") == "snakes_on_a_plane"
    assert to_snake_case("SnakesOnAPlane") == "snakes_on_a_plane"
    assert to_snake_case("SnakesOnA_Plane") == "snakes_on_a__plane"
    assert to_snake_case("snakes_on_a_plane") == "snakes_on_a_plane"
    assert to_snake_case("snakes_on_a__plane") == "snakes_on_a__plane"
    assert to_snake_case("IPhoneHysteria") == "i_phone_hysteria"
    assert to_snake_case("iPhoneHysteria") == "i_phone_hysteria"


def test_camel_case():
    assert to_camel_case("snakes_on_a_plane") == "snakesOnAPlane"
    assert to_camel_case("snakes_on_a__plane") == "snakesOnA_Plane"
    assert to_camel_case("i_phone_hysteria") == "iPhoneHysteria"
    assert to_camel_case("field_i18n") == "fieldI18n"



from ...types import ObjectType, Schema, String, NonNull


class Query(ObjectType):
    hello = String(input=NonNull(String))

    def resolve_hello(self, info, input):
        if input == "nothing":
            return None
        return f"Hello {input}!"


schema = Schema(query=Query)


def test_required_input_provided():
    """
    Test that a required argument works when provided.
    """
    input_value = "Potato"
    result = schema.execute('{ hello(input: "%s") }' % input_value)
    assert not result.errors
    assert result.data == {"hello": "Hello Potato!"}


def test_required_input_missing():
    """
    Test that a required argument raised an error if not provided.
    """
    result = schema.execute("{ hello }")
    assert result.errors
    assert len(result.errors) == 1
    assert (
        result.errors[0].message
        == "Field 'hello' argument 'input' of type 'String!' is required, but it was not provided."
    )



from functools import partial

from ..dynamic import Dynamic
from ..scalars import String
from ..structures import List, NonNull


def test_dynamic():
    dynamic = Dynamic(lambda: String)
    assert dynamic.get_type() == String
    assert str(dynamic.get_type()) == "String"


def test_nonnull():
    dynamic = Dynamic(lambda: NonNull(String))
    assert dynamic.get_type().of_type == String
    assert str(dynamic.get_type()) == "String!"


def test_list():
    dynamic = Dynamic(lambda: List(String))
    assert dynamic.get_type().of_type == String
    assert str(dynamic.get_type()) == "[String]"


def test_list_non_null():
    dynamic = Dynamic(lambda: List(NonNull(String)))
    assert dynamic.get_type().of_type.of_type == String
    assert str(dynamic.get_type()) == "[String!]"


def test_partial():
    def __type(_type):
        return _type

    dynamic = Dynamic(partial(__type, String))
    assert dynamic.get_type() == String
    assert str(dynamic.get_type()) == "String"



import graphene


class GeoInput(graphene.InputObjectType):
    lat = graphene.Float(required=True)
    lng = graphene.Float(required=True)

    @property
    def latlng(self):
        return f"({self.lat},{self.lng})"


class Address(graphene.ObjectType):
    latlng = graphene.String()


class Query(graphene.ObjectType):
    address = graphene.Field(Address, geo=GeoInput(required=True))

    def resolve_address(root, info, geo):
        return Address(latlng=geo.latlng)


class CreateAddress(graphene.Mutation):
    class Arguments:
        geo = GeoInput(required=True)

    Output = Address

    def mutate(root, info, geo):
        return Address(latlng=geo.latlng)


class Mutation(graphene.ObjectType):
    create_address = CreateAddress.Field()


schema = graphene.Schema(query=Query, mutation=Mutation)
query = """
    query something{
      address(geo: {lat:32.2, lng:12}) {
        latlng
      }
    }
"""
mutation = """
    mutation addAddress{
      createAddress(geo: {lat:32.2, lng:12}) {
        latlng
      }
    }
"""


def test_query():
    result = schema.execute(query)
    assert not result.errors
    assert result.data == {"address": {"latlng": "(32.2,12.0)"}}


def test_mutation():
    result = schema.execute(mutation)
    assert not result.errors
    assert result.data == {"createAddress": {"latlng": "(32.2,12.0)"}}


if __name__ == "__main__":
    result = schema.execute(query)
    print(result.data["address"]["latlng"])



from enum import Enum as PyEnum

from graphene.utils.subclass_with_meta import SubclassWithMeta_Meta

from .base import BaseOptions, BaseType
from .unmountedtype import UnmountedType


def eq_enum(self, other):
    if isinstance(other, self.__class__):
        return self is other
    return self.value is other


def hash_enum(self):
    return hash(self.name)


EnumType = type(PyEnum)


class EnumOptions(BaseOptions):
    enum = None  # type: Enum
    deprecation_reason = None


class EnumMeta(SubclassWithMeta_Meta):
    def __new__(cls, name_, bases, classdict, **options):
        enum_members = dict(classdict, __eq__=eq_enum, __hash__=hash_enum)
        # We remove the Meta attribute from the class to not collide
        # with the enum values.
        enum_members.pop("Meta", None)
        enum = PyEnum(cls.__name__, enum_members)
        obj = SubclassWithMeta_Meta.__new__(
            cls, name_, bases, dict(classdict, __enum__=enum), **options
        )
        globals()[name_] = obj.__enum__
        return obj

    def get(cls, value):
        return cls._meta.enum(value)

    def __getitem__(cls, value):
        return cls._meta.enum[value]

    def __prepare__(name, bases, **kwargs):  # noqa: N805
        return {}

    def __call__(cls, *args, **kwargs):  # noqa: N805
        if cls is Enum:
            description = kwargs.pop("description", None)
            deprecation_reason = kwargs.pop("deprecation_reason", None)
            return cls.from_enum(
                PyEnum(*args, **kwargs),
                description=description,
                deprecation_reason=deprecation_reason,
            )
        return super(EnumMeta, cls).__call__(*args, **kwargs)
        # return cls._meta.enum(*args, **kwargs)

    def __iter__(cls):
        return cls._meta.enum.__iter__()

    def from_enum(cls, enum, name=None, description=None, deprecation_reason=None):  # noqa: N805
        name = name or enum.__name__
        description = description or enum.__doc__ or "An enumeration."
        meta_dict = {
            "enum": enum,
            "description": description,
            "deprecation_reason": deprecation_reason,
        }
        meta_class = type("Meta", (object,), meta_dict)
        return type(name, (Enum,), {"Meta": meta_class})


class Enum(UnmountedType, BaseType, metaclass=EnumMeta):
    """
    Enum type definition

    Defines a static set of values that can be provided as a Field, Argument or InputField.

    .. code:: python

        from graphene import Enum

        class NameFormat(Enum):
            FIRST_LAST = "first_last"
            LAST_FIRST = "last_first"

    Meta:
        enum (optional, Enum): Python enum to use as a base for GraphQL Enum.

        name (optional, str): Name of the GraphQL type (must be unique in schema). Defaults to class
            name.
        description (optional, str): Description of the GraphQL type in the schema. Defaults to class
            docstring.
        deprecation_reason (optional, str): Setting this value indicates that the enum is
            depreciated and may provide instruction or reason on how for clients to proceed.
    """

    @classmethod
    def __init_subclass_with_meta__(cls, enum=None, _meta=None, **options):
        if not _meta:
            _meta = EnumOptions(cls)
        _meta.enum = enum or cls.__enum__
        _meta.deprecation_reason = options.pop("deprecation_reason", None)
        for key, value in _meta.enum.__members__.items():
            setattr(cls, key, value)

        super(Enum, cls).__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    def get_type(cls):
        """
        This function is called when the unmounted type (Enum instance)
        is mounted (as a Field, InputField or Argument)
        """
        return cls



import pytest

from ...types.base64 import Base64
from ...types.datetime import Date, DateTime
from ...types.decimal import Decimal
from ...types.generic import GenericScalar
from ...types.json import JSONString
from ...types.objecttype import ObjectType
from ...types.scalars import ID, BigInt, Boolean, Float, Int, String
from ...types.schema import Schema
from ...types.uuid import UUID


@pytest.mark.parametrize(
    "input_type,input_value",
    [
        (Date, '"2022-02-02"'),
        (GenericScalar, '"foo"'),
        (Int, "1"),
        (BigInt, "12345678901234567890"),
        (Float, "1.1"),
        (String, '"foo"'),
        (Boolean, "true"),
        (ID, "1"),
        (DateTime, '"2022-02-02T11:11:11"'),
        (UUID, '"cbebbc62-758e-4f75-a890-bc73b5017d81"'),
        (Decimal, '"1.1"'),
        (JSONString, '"{\\"key\\":\\"foo\\",\\"value\\":\\"bar\\"}"'),
        (Base64, '"Q2hlbG8gd29ycmxkCg=="'),
    ],
)
def test_parse_literal_with_variables(input_type, input_value):
    # input_b needs to be evaluated as literal while the variable dict for
    # input_a is passed along.

    class Query(ObjectType):
        generic = GenericScalar(input_a=GenericScalar(), input_b=input_type())

        def resolve_generic(self, info, input_a=None, input_b=None):
            return input

    schema = Schema(query=Query)

    query = f"""
        query Test($a: GenericScalar){{
            generic(inputA: $a, inputB: {input_value})
        }}
    """
    result = schema.execute(
        query,
        variables={"a": "bar"},
    )
    assert not result.errors



from ..generic import GenericScalar
from ..objecttype import ObjectType
from ..schema import Schema


class Query(ObjectType):
    generic = GenericScalar(input=GenericScalar())

    def resolve_generic(self, info, input=None):
        return input


schema = Schema(query=Query)


def test_generic_query_variable():
    for generic_value in [
        1,
        1.1,
        True,
        "str",
        [1, 2, 3],
        [1.1, 2.2, 3.3],
        [True, False],
        ["str1", "str2"],
        {"key_a": "a", "key_b": "b"},
        {
            "int": 1,
            "float": 1.1,
            "boolean": True,
            "string": "str",
            "int_list": [1, 2, 3],
            "float_list": [1.1, 2.2, 3.3],
            "boolean_list": [True, False],
            "string_list": ["str1", "str2"],
            "nested_dict": {"key_a": "a", "key_b": "b"},
        },
        None,
    ]:
        result = schema.execute(
            """query Test($generic: GenericScalar){ generic(input: $generic) }""",
            variables={"generic": generic_value},
        )
        assert not result.errors
        assert result.data == {"generic": generic_value}


def test_generic_parse_literal_query():
    result = schema.execute(
        """
        query {
            generic(input: {
                int: 1,
                float: 1.1
                boolean: true,
                string: "str",
                int_list: [1, 2, 3],
                float_list: [1.1, 2.2, 3.3],
                boolean_list: [true, false]
                string_list: ["str1", "str2"],
                nested_dict: {
                    key_a: "a",
                    key_b: "b"
                },
                empty_key: undefined
            })
        }
        """
    )
    assert not result.errors
    assert result.data == {
        "generic": {
            "int": 1,
            "float": 1.1,
            "boolean": True,
            "string": "str",
            "int_list": [1, 2, 3],
            "float_list": [1.1, 2.2, 3.3],
            "boolean_list": [True, False],
            "string_list": ["str1", "str2"],
            "nested_dict": {"key_a": "a", "key_b": "b"},
            "empty_key": None,
        }
    }



from textwrap import dedent

from pytest import raises

from graphql.type import GraphQLObjectType, GraphQLSchema

from ..field import Field
from ..objecttype import ObjectType
from ..scalars import String
from ..schema import Schema


class MyOtherType(ObjectType):
    field = String()


class Query(ObjectType):
    inner = Field(MyOtherType)


def test_schema():
    schema = Schema(Query)
    graphql_schema = schema.graphql_schema
    assert isinstance(graphql_schema, GraphQLSchema)
    query_type = graphql_schema.query_type
    assert isinstance(query_type, GraphQLObjectType)
    assert query_type.name == "Query"
    assert query_type.graphene_type is Query


def test_schema_get_type():
    schema = Schema(Query)
    assert schema.Query == Query
    assert schema.MyOtherType == MyOtherType


def test_schema_get_type_error():
    schema = Schema(Query)
    with raises(AttributeError) as exc_info:
        schema.X

    assert str(exc_info.value) == 'Type "X" not found in the Schema'


def test_schema_str():
    schema = Schema(Query)
    assert (
        str(schema).strip()
        == dedent(
            """
        type Query {
          inner: MyOtherType
        }

        type MyOtherType {
          field: String
        }
        """
        ).strip()
    )


def test_schema_introspect():
    schema = Schema(Query)
    assert "__schema" in schema.introspect()


def test_schema_requires_query_type():
    schema = Schema()
    result = schema.execute("query {}")

    assert len(result.errors) == 1
    error = result.errors[0]
    assert error.message == "Query root type must be provided."



import datetime
import graphene
from graphene import relay
from graphene.types.resolver import dict_resolver

from ..deduplicator import deflate


def test_does_not_modify_object_without_typename_and_id():
    response = {"foo": "bar"}

    deflated_response = deflate(response)
    assert deflated_response == {"foo": "bar"}


def test_does_not_modify_first_instance_of_an_object():
    response = {
        "data": [
            {"__typename": "foo", "id": 1, "name": "foo"},
            {"__typename": "foo", "id": 1, "name": "foo"},
        ]
    }

    deflated_response = deflate(response)

    assert deflated_response == {
        "data": [
            {"__typename": "foo", "id": 1, "name": "foo"},
            {"__typename": "foo", "id": 1},
        ]
    }


def test_does_not_modify_first_instance_of_an_object_nested():
    response = {
        "data": [
            {
                "__typename": "foo",
                "bar1": {"__typename": "bar", "id": 1, "name": "bar"},
                "bar2": {"__typename": "bar", "id": 1, "name": "bar"},
                "id": 1,
            },
            {
                "__typename": "foo",
                "bar1": {"__typename": "bar", "id": 1, "name": "bar"},
                "bar2": {"__typename": "bar", "id": 1, "name": "bar"},
                "id": 2,
            },
        ]
    }

    deflated_response = deflate(response)

    assert deflated_response == {
        "data": [
            {
                "__typename": "foo",
                "bar1": {"__typename": "bar", "id": 1, "name": "bar"},
                "bar2": {"__typename": "bar", "id": 1, "name": "bar"},
                "id": 1,
            },
            {
                "__typename": "foo",
                "bar1": {"__typename": "bar", "id": 1},
                "bar2": {"__typename": "bar", "id": 1},
                "id": 2,
            },
        ]
    }


def test_does_not_modify_input():
    response = {
        "data": [
            {"__typename": "foo", "id": 1, "name": "foo"},
            {"__typename": "foo", "id": 1, "name": "foo"},
        ]
    }

    deflate(response)

    assert response == {
        "data": [
            {"__typename": "foo", "id": 1, "name": "foo"},
            {"__typename": "foo", "id": 1, "name": "foo"},
        ]
    }


TEST_DATA = {
    "events": [
        {"id": "568", "date": datetime.date(2017, 5, 19), "movie": "1198359"},
        {"id": "234", "date": datetime.date(2017, 5, 20), "movie": "1198359"},
    ],
    "movies": {
        "1198359": {
            "id": "1198359",
            "name": "King Arthur: Legend of the Sword",
            "synopsis": (
                "When the child Arthur's father is murdered, Vortigern, "
                "Arthur's uncle, seizes the crown. Robbed of his birthright and "
                "with no idea who he truly is..."
            ),
        }
    },
}


def test_example_end_to_end():
    class Movie(graphene.ObjectType):
        class Meta:
            interfaces = (relay.Node,)
            default_resolver = dict_resolver

        name = graphene.String(required=True)
        synopsis = graphene.String(required=True)

    class Event(graphene.ObjectType):
        class Meta:
            interfaces = (relay.Node,)
            default_resolver = dict_resolver

        movie = graphene.Field(Movie, required=True)
        date = graphene.types.datetime.Date(required=True)

        def resolve_movie(event, info):
            return TEST_DATA["movies"][event["movie"]]

    class Query(graphene.ObjectType):
        events = graphene.List(graphene.NonNull(Event), required=True)

        def resolve_events(_, info):
            return TEST_DATA["events"]

    schema = graphene.Schema(query=Query)
    query = """\
        {
            events {
                __typename
                id
                date
                movie {
                    __typename
                    id
                    name
                    synopsis
                }
            }
        }
    """
    result = schema.execute(query)
    assert not result.errors

    data = deflate(result.data)
    assert data == {
        "events": [
            {
                "__typename": "Event",
                "id": "RXZlbnQ6NTY4",
                "date": "2017-05-19",
                "movie": {
                    "__typename": "Movie",
                    "id": "TW92aWU6MTE5ODM1OQ==",
                    "name": "King Arthur: Legend of the Sword",
                    "synopsis": (
                        "When the child Arthur's father is murdered, Vortigern, "
                        "Arthur's uncle, seizes the crown. Robbed of his birthright and "
                        "with no idea who he truly is..."
                    ),
                },
            },
            {
                "__typename": "Event",
                "id": "RXZlbnQ6MjM0",
                "date": "2017-05-20",
                "movie": {"__typename": "Movie", "id": "TW92aWU6MTE5ODM1OQ=="},
            },
        ]
    }



from typing import TYPE_CHECKING

from .base import BaseOptions, BaseType
from .field import Field
from .utils import yank_fields_from_attrs

# For static type checking with type checker
if TYPE_CHECKING:
    from typing import Dict, Iterable, Type  # NOQA


class InterfaceOptions(BaseOptions):
    fields = None  # type: Dict[str, Field]
    interfaces = ()  # type: Iterable[Type[Interface]]


class Interface(BaseType):
    """
    Interface Type Definition

    When a field can return one of a heterogeneous set of types, a Interface type
    is used to describe what types are possible, what fields are in common across
    all types, as well as a function to determine which type is actually used
    when the field is resolved.

    .. code:: python

        from graphene import Interface, String

        class HasAddress(Interface):
            class Meta:
                description = "Address fields"

            address1 = String()
            address2 = String()

    If a field returns an Interface Type, the ambiguous type of the object can be determined using
    ``resolve_type`` on Interface and an ObjectType with ``Meta.possible_types`` or ``is_type_of``.

    Meta:
        name (str): Name of the GraphQL type (must be unique in schema). Defaults to class
            name.
        description (str): Description of the GraphQL type in the schema. Defaults to class
            docstring.
        fields (Dict[str, graphene.Field]): Dictionary of field name to Field. Not recommended to
            use (prefer class attributes).
    """

    @classmethod
    def __init_subclass_with_meta__(cls, _meta=None, interfaces=(), **options):
        if not _meta:
            _meta = InterfaceOptions(cls)

        fields = {}
        for base in reversed(cls.__mro__):
            fields.update(yank_fields_from_attrs(base.__dict__, _as=Field))

        if _meta.fields:
            _meta.fields.update(fields)
        else:
            _meta.fields = fields

        if not _meta.interfaces:
            _meta.interfaces = interfaces

        super(Interface, cls).__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    def resolve_type(cls, instance, info):
        from .objecttype import ObjectType

        if isinstance(instance, ObjectType):
            return type(instance)

    def __init__(self, *args, **kwargs):
        raise Exception("An Interface cannot be initialized")



data = {}


def setup():
    global data

    from .schema import Ship, Faction

    xwing = Ship(id="1", name="X-Wing")

    ywing = Ship(id="2", name="Y-Wing")

    awing = Ship(id="3", name="A-Wing")

    # Yeah, technically it's Corellian. But it flew in the service of the rebels,
    # so for the purposes of this demo it's a rebel ship.
    falcon = Ship(id="4", name="Millennium Falcon")

    homeOne = Ship(id="5", name="Home One")

    tieFighter = Ship(id="6", name="TIE Fighter")

    tieInterceptor = Ship(id="7", name="TIE Interceptor")

    executor = Ship(id="8", name="Executor")

    rebels = Faction(
        id="1", name="Alliance to Restore the Republic", ships=["1", "2", "3", "4", "5"]
    )

    empire = Faction(id="2", name="Galactic Empire", ships=["6", "7", "8"])

    data = {
        "Faction": {"1": rebels, "2": empire},
        "Ship": {
            "1": xwing,
            "2": ywing,
            "3": awing,
            "4": falcon,
            "5": homeOne,
            "6": tieFighter,
            "7": tieInterceptor,
            "8": executor,
        },
    }


def create_ship(ship_name, faction_id):
    from .schema import Ship

    next_ship = len(data["Ship"].keys()) + 1
    new_ship = Ship(id=str(next_ship), name=ship_name)
    data["Ship"][new_ship.id] = new_ship
    data["Faction"][faction_id].ships.append(new_ship.id)
    return new_ship


def get_ship(_id):
    return data["Ship"][_id]


def get_faction(_id):
    return data["Faction"][_id]


def get_rebels():
    return get_faction("1")


def get_empire():
    return get_faction("2")



import graphene


class User(graphene.ObjectType):
    id = graphene.ID()
    name = graphene.String()


class Query(graphene.ObjectType):
    me = graphene.Field(User)

    def resolve_me(root, info):
        return info.context["user"]


schema = graphene.Schema(query=Query)
query = """
    query something{
      me {
        id
        name
      }
    }
"""


def test_query():
    result = schema.execute(query, context={"user": User(id="1", name="Syrus")})
    assert not result.errors
    assert result.data == {"me": {"id": "1", "name": "Syrus"}}


if __name__ == "__main__":
    result = schema.execute(query, context={"user": User(id="X", name="Console")})
    print(result.data["me"])



import textwrap

from graphene.test import Client

from ..data import setup
from ..schema import schema

setup()

client = Client(schema)


def test_str_schema():
    assert str(schema).strip() == textwrap.dedent(
        '''\
        type Query {
          rebels: Faction
          empire: Faction
          node(
            """The ID of the object"""
            id: ID!
          ): Node
        }

        """A faction in the Star Wars saga"""
        type Faction implements Node {
          """The ID of the object"""
          id: ID!

          """The name of the faction."""
          name: String

          """The ships used by the faction."""
          ships(before: String, after: String, first: Int, last: Int): ShipConnection
        }

        """An object with an ID"""
        interface Node {
          """The ID of the object"""
          id: ID!
        }

        type ShipConnection {
          """Pagination data for this connection."""
          pageInfo: PageInfo!

          """Contains the nodes in this connection."""
          edges: [ShipEdge]!
        }

        """
        The Relay compliant `PageInfo` type, containing data necessary to paginate this connection.
        """
        type PageInfo {
          """When paginating forwards, are there more items?"""
          hasNextPage: Boolean!

          """When paginating backwards, are there more items?"""
          hasPreviousPage: Boolean!

          """When paginating backwards, the cursor to continue."""
          startCursor: String

          """When paginating forwards, the cursor to continue."""
          endCursor: String
        }

        """A Relay edge containing a `Ship` and its cursor."""
        type ShipEdge {
          """The item at the end of the edge"""
          node: Ship

          """A cursor for use in pagination"""
          cursor: String!
        }

        """A ship in the Star Wars saga"""
        type Ship implements Node {
          """The ID of the object"""
          id: ID!

          """The name of the ship."""
          name: String
        }

        type Mutation {
          introduceShip(input: IntroduceShipInput!): IntroduceShipPayload
        }

        type IntroduceShipPayload {
          ship: Ship
          faction: Faction
          clientMutationId: String
        }

        input IntroduceShipInput {
          shipName: String!
          factionId: String!
          clientMutationId: String
        }'''
    )


def test_correctly_fetches_id_name_rebels():
    result = client.execute("""
      query RebelsQuery {
        rebels {
          id
          name
        }
      }
    """)
    assert result == {
        "data": {
            "rebels": {"id": "RmFjdGlvbjox", "name": "Alliance to Restore the Republic"}
        }
    }


def test_correctly_refetches_rebels():
    result = client.execute("""
      query RebelsRefetchQuery {
        node(id: "RmFjdGlvbjox") {
          id
          ... on Faction {
            name
          }
        }
      }
    """)
    assert result == {
        "data": {
            "node": {"id": "RmFjdGlvbjox", "name": "Alliance to Restore the Republic"}
        }
    }


def test_correctly_fetches_id_name_empire():
    result = client.execute("""
      query EmpireQuery {
        empire {
          id
          name
        }
      }
    """)
    assert result == {
        "data": {"empire": {"id": "RmFjdGlvbjoy", "name": "Galactic Empire"}}
    }


def test_correctly_refetches_empire():
    result = client.execute("""
      query EmpireRefetchQuery {
        node(id: "RmFjdGlvbjoy") {
          id
          ... on Faction {
            name
          }
        }
      }
    """)
    assert result == {
        "data": {"node": {"id": "RmFjdGlvbjoy", "name": "Galactic Empire"}}
    }


def test_correctly_refetches_xwing():
    result = client.execute("""
      query XWingRefetchQuery {
        node(id: "U2hpcDox") {
          id
          ... on Ship {
            name
          }
        }
      }
    """)
    assert result == {"data": {"node": {"id": "U2hpcDox", "name": "X-Wing"}}}



from graphene.test import Client

from ..data import setup
from ..schema import schema

setup()

client = Client(schema)


def test_correct_fetch_first_ship_rebels():
    result = client.execute("""
        query RebelsShipsQuery {
          rebels {
            name,
            ships(first: 1) {
              pageInfo {
                startCursor
                endCursor
                hasNextPage
                hasPreviousPage
              }
              edges {
                cursor
                node {
                  name
                }
              }
            }
          }
        }
    """)
    assert result == {
        "data": {
            "rebels": {
                "name": "Alliance to Restore the Republic",
                "ships": {
                    "pageInfo": {
                        "startCursor": "YXJyYXljb25uZWN0aW9uOjA=",
                        "endCursor": "YXJyYXljb25uZWN0aW9uOjA=",
                        "hasNextPage": True,
                        "hasPreviousPage": False,
                    },
                    "edges": [
                        {
                            "cursor": "YXJyYXljb25uZWN0aW9uOjA=",
                            "node": {"name": "X-Wing"},
                        }
                    ],
                },
            }
        }
    }



from itertools import chain
from graphql import Undefined

from .dynamic import Dynamic
from .mountedtype import MountedType
from .structures import NonNull
from .utils import get_type


class Argument(MountedType):
    """
    Makes an Argument available on a Field in the GraphQL schema.

    Arguments will be parsed and provided to resolver methods for fields as keyword arguments.

    All ``arg`` and ``**extra_args`` for a ``graphene.Field`` are implicitly mounted as Argument
    using the below parameters.

    .. code:: python

        from graphene import String, Boolean, Argument

        age = String(
            # Boolean implicitly mounted as Argument
            dog_years=Boolean(description="convert to dog years"),
            # Boolean explicitly mounted as Argument
            decades=Argument(Boolean, default_value=False),
        )

    args:
        type (class for a graphene.UnmountedType): must be a class (not an instance) of an
            unmounted graphene type (ex. scalar or object) which is used for the type of this
            argument in the GraphQL schema.
        required (optional, bool): indicates this argument as not null in the graphql schema. Same behavior
            as graphene.NonNull. Default False.
        name (optional, str): the name of the GraphQL argument. Defaults to parameter name.
        description (optional, str): the description of the GraphQL argument in the schema.
        default_value (optional, Any): The value to be provided if the user does not set this argument in
            the operation.
        deprecation_reason (optional, str): Setting this value indicates that the argument is
            depreciated and may provide instruction or reason on how for clients to proceed. Cannot be
            set if the argument is required (see spec).
    """

    def __init__(
        self,
        type_,
        default_value=Undefined,
        deprecation_reason=None,
        description=None,
        name=None,
        required=False,
        _creation_counter=None,
    ):
        super(Argument, self).__init__(_creation_counter=_creation_counter)

        if required:
            assert (
                deprecation_reason is None
            ), f"Argument {name} is required, cannot deprecate it."
            type_ = NonNull(type_)

        self.name = name
        self._type = type_
        self.default_value = default_value
        self.description = description
        self.deprecation_reason = deprecation_reason

    @property
    def type(self):
        return get_type(self._type)

    def __eq__(self, other):
        return isinstance(other, Argument) and (
            self.name == other.name
            and self.type == other.type
            and self.default_value == other.default_value
            and self.description == other.description
            and self.deprecation_reason == other.deprecation_reason
        )


def to_arguments(args, extra_args=None):
    from .unmountedtype import UnmountedType
    from .field import Field
    from .inputfield import InputField

    if extra_args:
        extra_args = sorted(extra_args.items(), key=lambda f: f[1])
    else:
        extra_args = []
    iter_arguments = chain(args.items(), extra_args)
    arguments = {}
    for default_name, arg in iter_arguments:
        if isinstance(arg, Dynamic):
            arg = arg.get_type()
            if arg is None:
                # If the Dynamic type returned None
                # then we skip the Argument
                continue

        if isinstance(arg, UnmountedType):
            arg = Argument.mounted(arg)

        if isinstance(arg, (InputField, Field)):
            raise ValueError(
                f"Expected {default_name} to be Argument, "
                f"but received {type(arg).__name__}. Try using Argument({arg.type})."
            )

        if not isinstance(arg, Argument):
            raise ValueError(f'Unknown argument "{default_name}".')

        arg_name = default_name or arg.name
        assert (
            arg_name not in arguments
        ), f'More than one Argument have same name "{arg_name}".'
        arguments[arg_name] = arg

    return arguments



import graphene

from .data import get_character, get_droid, get_hero, get_human


class Episode(graphene.Enum):
    NEWHOPE = 4
    EMPIRE = 5
    JEDI = 6


class Character(graphene.Interface):
    id = graphene.ID()
    name = graphene.String()
    friends = graphene.List(lambda: Character)
    appears_in = graphene.List(Episode)

    def resolve_friends(self, info):
        # The character friends is a list of strings
        return [get_character(f) for f in self.friends]


class Human(graphene.ObjectType):
    class Meta:
        interfaces = (Character,)

    home_planet = graphene.String()


class Droid(graphene.ObjectType):
    class Meta:
        interfaces = (Character,)

    primary_function = graphene.String()


class Query(graphene.ObjectType):
    hero = graphene.Field(Character, episode=Episode())
    human = graphene.Field(Human, id=graphene.String())
    droid = graphene.Field(Droid, id=graphene.String())

    def resolve_hero(root, info, episode=None):
        return get_hero(episode)

    def resolve_human(root, info, id):
        return get_human(id)

    def resolve_droid(root, info, id):
        return get_droid(id)


schema = graphene.Schema(query=Query)



from decimal import Decimal as _Decimal

from graphql import Undefined
from graphql.language.ast import StringValueNode, IntValueNode

from .scalars import Scalar


class Decimal(Scalar):
    """
    The `Decimal` scalar type represents a python Decimal.
    """

    @staticmethod
    def serialize(dec):
        if isinstance(dec, str):
            dec = _Decimal(dec)
        assert isinstance(
            dec, _Decimal
        ), f'Received not compatible Decimal "{repr(dec)}"'
        return str(dec)

    @classmethod
    def parse_literal(cls, node, _variables=None):
        if isinstance(node, (StringValueNode, IntValueNode)):
            return cls.parse_value(node.value)
        return Undefined

    @staticmethod
    def parse_value(value):
        try:
            return _Decimal(value)
        except Exception:
            return Undefined



from graphql import GraphQLResolveInfo as ResolveInfo

from .argument import Argument
from .base64 import Base64
from .context import Context
from .datetime import Date, DateTime, Time
from .decimal import Decimal
from .dynamic import Dynamic
from .enum import Enum
from .field import Field
from .inputfield import InputField
from .inputobjecttype import InputObjectType
from .interface import Interface
from .json import JSONString
from .mutation import Mutation
from .objecttype import ObjectType
from .scalars import ID, BigInt, Boolean, Float, Int, Scalar, String
from .schema import Schema
from .structures import List, NonNull
from .union import Union
from .uuid import UUID

__all__ = [
    "Argument",
    "Base64",
    "BigInt",
    "Boolean",
    "Context",
    "Date",
    "DateTime",
    "Decimal",
    "Dynamic",
    "Enum",
    "Field",
    "Float",
    "ID",
    "InputField",
    "InputObjectType",
    "Int",
    "Interface",
    "JSONString",
    "List",
    "Mutation",
    "NonNull",
    "ObjectType",
    "ResolveInfo",
    "Scalar",
    "Schema",
    "String",
    "Time",
    "UUID",
    "Union",
]



import graphene


class Patron(graphene.ObjectType):
    id = graphene.ID()
    name = graphene.String()
    age = graphene.Int()


class Query(graphene.ObjectType):
    patron = graphene.Field(Patron)

    def resolve_patron(root, info):
        return Patron(id=1, name="Syrus", age=27)


schema = graphene.Schema(query=Query)
query = """
    query something{
      patron {
        id
        name
        age
      }
    }
"""


def test_query():
    result = schema.execute(query)
    assert not result.errors
    assert result.data == {"patron": {"id": "1", "name": "Syrus", "age": 27}}


if __name__ == "__main__":
    result = schema.execute(query)
    print(result.data["patron"])



import ast
import codecs
import re
import sys

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

_version_re = re.compile(r"VERSION\s+=\s+(.*)")

with open("graphene/__init__.py", "rb") as f:
    version = ast.literal_eval(_version_re.search(f.read().decode("utf-8")).group(1))

path_copy = sys.path[:]

sys.path.append("graphene")
try:
    from pyutils.version import get_version

    version = get_version(version)
except Exception:
    version = ".".join([str(v) for v in version])

sys.path[:] = path_copy


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


tests_require = [
    "pytest>=8,<9",
    "pytest-benchmark>=4,<5",
    "pytest-cov>=5,<6",
    "pytest-mock>=3,<4",
    "pytest-asyncio>=0.16,<2",
    "coveralls>=3.3,<5",
]

dev_requires = [
    "ruff==0.5.0",
    "types-python-dateutil>=2.8.1,<3",
    "mypy>=1.10,<2",
] + tests_require

setup(
    name="graphene",
    version=version,
    description="GraphQL Framework for Python",
    long_description=codecs.open(
        "README.md", "r", encoding="ascii", errors="replace"
    ).read(),
    long_description_content_type="text/markdown",
    url="https://github.com/graphql-python/graphene",
    author="Syrus Akbary",
    author_email="me@syrusakbary.com",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
    keywords="api graphql protocol rest relay graphene",
    packages=find_packages(exclude=["examples*"]),
    install_requires=[
        "graphql-core>=3.1,<3.3",
        "graphql-relay>=3.1,<3.3",
        "python-dateutil>=2.7.0,<3",
        "typing-extensions>=4.7.1,<5",
    ],
    tests_require=tests_require,
    extras_require={"test": tests_require, "dev": dev_requires},
    cmdclass={"test": PyTest},
)



from asyncio import gather
from collections import namedtuple
from functools import partial
from unittest.mock import Mock

from graphene.utils.dataloader import DataLoader
from pytest import mark, raises

from graphene import ObjectType, String, Schema, Field, List

CHARACTERS = {
    "1": {"name": "Luke Skywalker", "sibling": "3"},
    "2": {"name": "Darth Vader", "sibling": None},
    "3": {"name": "Leia Organa", "sibling": "1"},
}

get_character = Mock(side_effect=lambda character_id: CHARACTERS[character_id])


class CharacterType(ObjectType):
    name = String()
    sibling = Field(lambda: CharacterType)

    async def resolve_sibling(character, info):
        if character["sibling"]:
            return await info.context.character_loader.load(character["sibling"])
        return None


class Query(ObjectType):
    skywalker_family = List(CharacterType)

    async def resolve_skywalker_family(_, info):
        return await info.context.character_loader.load_many(["1", "2", "3"])


mock_batch_load_fn = Mock(
    side_effect=lambda character_ids: [get_character(id) for id in character_ids]
)


class CharacterLoader(DataLoader):
    async def batch_load_fn(self, character_ids):
        return mock_batch_load_fn(character_ids)


Context = namedtuple("Context", "character_loader")


@mark.asyncio
async def test_basic_dataloader():
    schema = Schema(query=Query)

    character_loader = CharacterLoader()
    context = Context(character_loader=character_loader)

    query = """
        {
            skywalkerFamily {
                name
                sibling {
                    name
                }
            }
        }
    """

    result = await schema.execute_async(query, context=context)

    assert not result.errors
    assert result.data == {
        "skywalkerFamily": [
            {"name": "Luke Skywalker", "sibling": {"name": "Leia Organa"}},
            {"name": "Darth Vader", "sibling": None},
            {"name": "Leia Organa", "sibling": {"name": "Luke Skywalker"}},
        ]
    }

    assert mock_batch_load_fn.call_count == 1
    assert get_character.call_count == 3


def id_loader(**options):
    load_calls = []

    async def default_resolve(x):
        return x

    resolve = options.pop("resolve", default_resolve)

    async def fn(keys):
        load_calls.append(keys)
        return await resolve(keys)
        # return keys

    identity_loader = DataLoader(fn, **options)
    return identity_loader, load_calls


@mark.asyncio
async def test_build_a_simple_data_loader():
    async def call_fn(keys):
        return keys

    identity_loader = DataLoader(call_fn)

    promise1 = identity_loader.load(1)

    value1 = await promise1
    assert value1 == 1


@mark.asyncio
async def test_can_build_a_data_loader_from_a_partial():
    value_map = {1: "one"}

    async def call_fn(context, keys):
        return [context.get(key) for key in keys]

    partial_fn = partial(call_fn, value_map)
    identity_loader = DataLoader(partial_fn)

    promise1 = identity_loader.load(1)

    value1 = await promise1
    assert value1 == "one"


@mark.asyncio
async def test_supports_loading_multiple_keys_in_one_call():
    async def call_fn(keys):
        return keys

    identity_loader = DataLoader(call_fn)

    promise_all = identity_loader.load_many([1, 2])

    values = await promise_all
    assert values == [1, 2]

    promise_all = identity_loader.load_many([])

    values = await promise_all
    assert values == []


@mark.asyncio
async def test_batches_multiple_requests():
    identity_loader, load_calls = id_loader()

    promise1 = identity_loader.load(1)
    promise2 = identity_loader.load(2)

    p = gather(promise1, promise2)

    value1, value2 = await p

    assert value1 == 1
    assert value2 == 2

    assert load_calls == [[1, 2]]


@mark.asyncio
async def test_batches_multiple_requests_with_max_batch_sizes():
    identity_loader, load_calls = id_loader(max_batch_size=2)

    promise1 = identity_loader.load(1)
    promise2 = identity_loader.load(2)
    promise3 = identity_loader.load(3)

    p = gather(promise1, promise2, promise3)

    value1, value2, value3 = await p

    assert value1 == 1
    assert value2 == 2
    assert value3 == 3

    assert load_calls == [[1, 2], [3]]


@mark.asyncio
async def test_coalesces_identical_requests():
    identity_loader, load_calls = id_loader()

    promise1 = identity_loader.load(1)
    promise2 = identity_loader.load(1)

    assert promise1 == promise2
    p = gather(promise1, promise2)

    value1, value2 = await p

    assert value1 == 1
    assert value2 == 1

    assert load_calls == [[1]]


@mark.asyncio
async def test_caches_repeated_requests():
    identity_loader, load_calls = id_loader()

    a, b = await gather(identity_loader.load("A"), identity_loader.load("B"))

    assert a == "A"
    assert b == "B"

    assert load_calls == [["A", "B"]]

    a2, c = await gather(identity_loader.load("A"), identity_loader.load("C"))

    assert a2 == "A"
    assert c == "C"

    assert load_calls == [["A", "B"], ["C"]]

    a3, b2, c2 = await gather(
        identity_loader.load("A"), identity_loader.load("B"), identity_loader.load("C")
    )

    assert a3 == "A"
    assert b2 == "B"
    assert c2 == "C"

    assert load_calls == [["A", "B"], ["C"]]


@mark.asyncio
async def test_clears_single_value_in_loader():
    identity_loader, load_calls = id_loader()

    a, b = await gather(identity_loader.load("A"), identity_loader.load("B"))

    assert a == "A"
    assert b == "B"

    assert load_calls == [["A", "B"]]

    identity_loader.clear("A")

    a2, b2 = await gather(identity_loader.load("A"), identity_loader.load("B"))

    assert a2 == "A"
    assert b2 == "B"

    assert load_calls == [["A", "B"], ["A"]]


@mark.asyncio
async def test_clears_all_values_in_loader():
    identity_loader, load_calls = id_loader()

    a, b = await gather(identity_loader.load("A"), identity_loader.load("B"))

    assert a == "A"
    assert b == "B"

    assert load_calls == [["A", "B"]]

    identity_loader.clear_all()

    a2, b2 = await gather(identity_loader.load("A"), identity_loader.load("B"))

    assert a2 == "A"
    assert b2 == "B"

    assert load_calls == [["A", "B"], ["A", "B"]]


@mark.asyncio
async def test_allows_priming_the_cache():
    identity_loader, load_calls = id_loader()

    identity_loader.prime("A", "A")

    a, b = await gather(identity_loader.load("A"), identity_loader.load("B"))

    assert a == "A"
    assert b == "B"

    assert load_calls == [["B"]]


@mark.asyncio
async def test_does_not_prime_keys_that_already_exist():
    identity_loader, load_calls = id_loader()

    identity_loader.prime("A", "X")

    a1 = await identity_loader.load("A")
    b1 = await identity_loader.load("B")

    assert a1 == "X"
    assert b1 == "B"

    identity_loader.prime("A", "Y")
    identity_loader.prime("B", "Y")

    a2 = await identity_loader.load("A")
    b2 = await identity_loader.load("B")

    assert a2 == "X"
    assert b2 == "B"

    assert load_calls == [["B"]]


# # Represents Errors
@mark.asyncio
async def test_resolves_to_error_to_indicate_failure():
    async def resolve(keys):
        mapped_keys = [
            key if key % 2 == 0 else Exception("Odd: {}".format(key)) for key in keys
        ]
        return mapped_keys

    even_loader, load_calls = id_loader(resolve=resolve)

    with raises(Exception) as exc_info:
        await even_loader.load(1)

    assert str(exc_info.value) == "Odd: 1"

    value2 = await even_loader.load(2)
    assert value2 == 2
    assert load_calls == [[1], [2]]


@mark.asyncio
async def test_can_represent_failures_and_successes_simultaneously():
    async def resolve(keys):
        mapped_keys = [
            key if key % 2 == 0 else Exception("Odd: {}".format(key)) for key in keys
        ]
        return mapped_keys

    even_loader, load_calls = id_loader(resolve=resolve)

    promise1 = even_loader.load(1)
    promise2 = even_loader.load(2)

    with raises(Exception) as exc_info:
        await promise1

    assert str(exc_info.value) == "Odd: 1"
    value2 = await promise2
    assert value2 == 2
    assert load_calls == [[1, 2]]


@mark.asyncio
async def test_caches_failed_fetches():
    async def resolve(keys):
        mapped_keys = [Exception("Error: {}".format(key)) for key in keys]
        return mapped_keys

    error_loader, load_calls = id_loader(resolve=resolve)

    with raises(Exception) as exc_info:
        await error_loader.load(1)

    assert str(exc_info.value) == "Error: 1"

    with raises(Exception) as exc_info:
        await error_loader.load(1)

    assert str(exc_info.value) == "Error: 1"

    assert load_calls == [[1]]


@mark.asyncio
async def test_caches_failed_fetches_2():
    identity_loader, load_calls = id_loader()

    identity_loader.prime(1, Exception("Error: 1"))

    with raises(Exception) as _:
        await identity_loader.load(1)

    assert load_calls == []


# It is resilient to job queue ordering
@mark.asyncio
async def test_batches_loads_occuring_within_promises():
    identity_loader, load_calls = id_loader()

    async def load_b_1():
        return await load_b_2()

    async def load_b_2():
        return await identity_loader.load("B")

    values = await gather(identity_loader.load("A"), load_b_1())

    assert values == ["A", "B"]

    assert load_calls == [["A", "B"]]


@mark.asyncio
async def test_catches_error_if_loader_resolver_fails():
    exc = Exception("AOH!")

    def do_resolve(x):
        raise exc

    a_loader, a_load_calls = id_loader(resolve=do_resolve)

    with raises(Exception) as exc_info:
        await a_loader.load("A1")

    assert exc_info.value == exc


@mark.asyncio
async def test_can_call_a_loader_from_a_loader():
    deep_loader, deep_load_calls = id_loader()
    a_loader, a_load_calls = id_loader(
        resolve=lambda keys: deep_loader.load(tuple(keys))
    )
    b_loader, b_load_calls = id_loader(
        resolve=lambda keys: deep_loader.load(tuple(keys))
    )

    a1, b1, a2, b2 = await gather(
        a_loader.load("A1"),
        b_loader.load("B1"),
        a_loader.load("A2"),
        b_loader.load("B2"),
    )

    assert a1 == "A1"
    assert b1 == "B1"
    assert a2 == "A2"
    assert b2 == "B2"

    assert a_load_calls == [["A1", "A2"]]
    assert b_load_calls == [["B1", "B2"]]
    assert deep_load_calls == [[("A1", "A2"), ("B1", "B2")]]


@mark.asyncio
async def test_dataloader_clear_with_missing_key_works():
    async def do_resolve(x):
        return x

    a_loader, a_load_calls = id_loader(resolve=do_resolve)
    assert a_loader.clear("A1") == a_loader



# https://github.com/graphql-python/graphene/issues/720
# InputObjectTypes overwrite the "fields" attribute of the provided
# _meta object, so even if dynamic fields are provided with a standard
# InputObjectTypeOptions, they are ignored.

import graphene


class MyInputClass(graphene.InputObjectType):
    @classmethod
    def __init_subclass_with_meta__(
        cls, container=None, _meta=None, fields=None, **options
    ):
        if _meta is None:
            _meta = graphene.types.inputobjecttype.InputObjectTypeOptions(cls)
        _meta.fields = fields
        super(MyInputClass, cls).__init_subclass_with_meta__(
            container=container, _meta=_meta, **options
        )


class MyInput(MyInputClass):
    class Meta:
        fields = dict(x=graphene.Field(graphene.Int))


class Query(graphene.ObjectType):
    myField = graphene.Field(graphene.String, input=graphene.Argument(MyInput))

    def resolve_myField(parent, info, input):
        return "ok"


def test_issue():
    query_string = """
    query myQuery {
      myField(input: {x: 1})
    }
    """

    schema = graphene.Schema(query=Query)
    result = schema.execute(query_string)

    assert not result.errors



from pytest import mark

from ..crunch import crunch


@mark.parametrize(
    "description,uncrunched,crunched",
    [
        ["number primitive", 0, [0]],
        ["boolean primitive", True, [True]],
        ["string primitive", "string", ["string"]],
        ["empty array", [], [[]]],
        ["single-item array", [None], [None, [0]]],
        [
            "multi-primitive all distinct array",
            [None, 0, True, "string"],
            [None, 0, True, "string", [0, 1, 2, 3]],
        ],
        [
            "multi-primitive repeated array",
            [True, True, True, True],
            [True, [0, 0, 0, 0]],
        ],
        ["one-level nested array", [[1, 2, 3]], [1, 2, 3, [0, 1, 2], [3]]],
        ["two-level nested array", [[[1, 2, 3]]], [1, 2, 3, [0, 1, 2], [3], [4]]],
        ["empty object", {}, [{}]],
        ["single-item object", {"a": None}, [None, {"a": 0}]],
        [
            "multi-item all distinct object",
            {"a": None, "b": 0, "c": True, "d": "string"},
            [None, 0, True, "string", {"a": 0, "b": 1, "c": 2, "d": 3}],
        ],
        [
            "multi-item repeated object",
            {"a": True, "b": True, "c": True, "d": True},
            [True, {"a": 0, "b": 0, "c": 0, "d": 0}],
        ],
        [
            "complex array",
            [{"a": True, "b": [1, 2, 3]}, [1, 2, 3]],
            [True, 1, 2, 3, [1, 2, 3], {"a": 0, "b": 4}, [5, 4]],
        ],
        [
            "complex object",
            {"a": True, "b": [1, 2, 3], "c": {"a": True, "b": [1, 2, 3]}},
            [True, 1, 2, 3, [1, 2, 3], {"a": 0, "b": 4}, {"a": 0, "b": 4, "c": 5}],
        ],
    ],
)
def test_crunch(description, uncrunched, crunched):
    assert crunch(uncrunched) == crunched



from ..utils.orderedtype import OrderedType
from .unmountedtype import UnmountedType


class MountedType(OrderedType):
    @classmethod
    def mounted(cls, unmounted):  # noqa: N802
        """
        Mount the UnmountedType instance
        """
        assert isinstance(
            unmounted, UnmountedType
        ), f"{cls.__name__} can't mount {repr(unmounted)}"

        return cls(
            unmounted.get_type(),
            *unmounted.args,
            _creation_counter=unmounted.creation_counter,
            **unmounted.kwargs,
        )



from typing import Type, Optional

from ..utils.subclass_with_meta import SubclassWithMeta, SubclassWithMeta_Meta
from ..utils.trim_docstring import trim_docstring


class BaseOptions:
    name: Optional[str] = None
    description: Optional[str] = None

    _frozen: bool = False

    def __init__(self, class_type: Type):
        self.class_type: Type = class_type

    def freeze(self):
        self._frozen = True

    def __setattr__(self, name, value):
        if not self._frozen:
            super(BaseOptions, self).__setattr__(name, value)
        else:
            raise Exception(f"Can't modify frozen Options {self}")

    def __repr__(self):
        return f"<{self.__class__.__name__} name={repr(self.name)}>"


BaseTypeMeta = SubclassWithMeta_Meta


class BaseType(SubclassWithMeta):
    @classmethod
    def create_type(cls, class_name, **options):
        return type(class_name, (cls,), {"Meta": options})

    @classmethod
    def __init_subclass_with_meta__(
        cls, name=None, description=None, _meta=None, **_kwargs
    ):
        assert "_meta" not in cls.__dict__, "Can't assign meta directly"
        if not _meta:
            return
        _meta.name = name or cls.__name__
        _meta.description = description or trim_docstring(cls.__doc__)
        _meta.freeze()
        cls._meta = _meta
        super(BaseType, cls).__init_subclass_with_meta__()



from pytest import raises

from ..field import Field
from ..interface import Interface
from ..objecttype import ObjectType
from ..scalars import String
from ..schema import Schema
from ..structures import NonNull
from ..unmountedtype import UnmountedType


class MyType(Interface):
    pass


class Container(ObjectType):
    field1 = Field(MyType)
    field2 = Field(MyType)


class MyInterface(Interface):
    ifield = Field(MyType)


class ContainerWithInterface(ObjectType):
    class Meta:
        interfaces = (MyInterface,)

    field1 = Field(MyType)
    field2 = Field(MyType)


class MyScalar(UnmountedType):
    def get_type(self):
        return MyType


def test_generate_objecttype():
    class MyObjectType(ObjectType):
        """Documentation"""

    assert MyObjectType._meta.name == "MyObjectType"
    assert MyObjectType._meta.description == "Documentation"
    assert MyObjectType._meta.interfaces == tuple()
    assert MyObjectType._meta.fields == {}
    assert (
        repr(MyObjectType)
        == "<MyObjectType meta=<ObjectTypeOptions name='MyObjectType'>>"
    )


def test_generate_objecttype_with_meta():
    class MyObjectType(ObjectType):
        class Meta:
            name = "MyOtherObjectType"
            description = "Documentation"
            interfaces = (MyType,)

    assert MyObjectType._meta.name == "MyOtherObjectType"
    assert MyObjectType._meta.description == "Documentation"
    assert MyObjectType._meta.interfaces == (MyType,)


def test_generate_lazy_objecttype():
    class MyObjectType(ObjectType):
        example = Field(lambda: InnerObjectType, required=True)

    class InnerObjectType(ObjectType):
        field = Field(MyType)

    assert MyObjectType._meta.name == "MyObjectType"
    example_field = MyObjectType._meta.fields["example"]
    assert isinstance(example_field.type, NonNull)
    assert example_field.type.of_type == InnerObjectType


def test_generate_objecttype_with_fields():
    class MyObjectType(ObjectType):
        field = Field(MyType)

    assert "field" in MyObjectType._meta.fields


def test_generate_objecttype_with_private_attributes():
    class MyObjectType(ObjectType):
        def __init__(self, _private_state=None, **kwargs):
            self._private_state = _private_state
            super().__init__(**kwargs)

        _private_state = None

    assert "_private_state" not in MyObjectType._meta.fields
    assert hasattr(MyObjectType, "_private_state")

    m = MyObjectType(_private_state="custom")
    assert m._private_state == "custom"

    with raises(TypeError):
        MyObjectType(_other_private_state="Wrong")


def test_ordered_fields_in_objecttype():
    class MyObjectType(ObjectType):
        b = Field(MyType)
        a = Field(MyType)
        field = MyScalar()
        asa = Field(MyType)

    assert list(MyObjectType._meta.fields) == ["b", "a", "field", "asa"]


def test_generate_objecttype_inherit_abstracttype():
    class MyAbstractType:
        field1 = MyScalar()

    class MyObjectType(ObjectType, MyAbstractType):
        field2 = MyScalar()

    assert MyObjectType._meta.description is None
    assert MyObjectType._meta.interfaces == ()
    assert MyObjectType._meta.name == "MyObjectType"
    assert list(MyObjectType._meta.fields) == ["field1", "field2"]
    assert list(map(type, MyObjectType._meta.fields.values())) == [Field, Field]


def test_generate_objecttype_inherit_abstracttype_reversed():
    class MyAbstractType:
        field1 = MyScalar()

    class MyObjectType(MyAbstractType, ObjectType):
        field2 = MyScalar()

    assert MyObjectType._meta.description is None
    assert MyObjectType._meta.interfaces == ()
    assert MyObjectType._meta.name == "MyObjectType"
    assert list(MyObjectType._meta.fields) == ["field1", "field2"]
    assert list(map(type, MyObjectType._meta.fields.values())) == [Field, Field]


def test_generate_objecttype_unmountedtype():
    class MyObjectType(ObjectType):
        field = MyScalar()

    assert "field" in MyObjectType._meta.fields
    assert isinstance(MyObjectType._meta.fields["field"], Field)


def test_parent_container_get_fields():
    assert list(Container._meta.fields) == ["field1", "field2"]


def test_parent_container_interface_get_fields():
    assert list(ContainerWithInterface._meta.fields) == ["ifield", "field1", "field2"]


def test_objecttype_as_container_only_args():
    container = Container("1", "2")
    assert container.field1 == "1"
    assert container.field2 == "2"


def test_objecttype_repr():
    container = Container("1", "2")
    assert repr(container) == "Container(field1='1', field2='2')"


def test_objecttype_eq():
    container1 = Container("1", "2")
    container2 = Container("1", "2")
    container3 = Container("2", "3")
    assert container1 == container1
    assert container1 == container2
    assert container2 != container3


def test_objecttype_as_container_args_kwargs():
    container = Container("1", field2="2")
    assert container.field1 == "1"
    assert container.field2 == "2"


def test_objecttype_as_container_few_kwargs():
    container = Container(field2="2")
    assert container.field2 == "2"


def test_objecttype_as_container_all_kwargs():
    container = Container(field1="1", field2="2")
    assert container.field1 == "1"
    assert container.field2 == "2"


def test_objecttype_as_container_extra_args():
    msg = r"__init__\(\) takes from 1 to 3 positional arguments but 4 were given"
    with raises(TypeError, match=msg):
        Container("1", "2", "3")  # type: ignore


def test_objecttype_as_container_invalid_kwargs():
    msg = r"__init__\(\) got an unexpected keyword argument 'unexisting_field'"
    with raises(TypeError, match=msg):
        Container(unexisting_field="3")  # type: ignore


def test_objecttype_container_benchmark(benchmark):
    @benchmark
    def create_objecttype():
        Container(field1="field1", field2="field2")


def test_generate_objecttype_description():
    class MyObjectType(ObjectType):
        """
        Documentation

        Documentation line 2
        """

    assert MyObjectType._meta.description == "Documentation\n\nDocumentation line 2"


def test_objecttype_with_possible_types():
    class MyObjectType(ObjectType):
        class Meta:
            possible_types = (dict,)

    assert MyObjectType._meta.possible_types == (dict,)


def test_objecttype_with_possible_types_and_is_type_of_should_raise():
    with raises(AssertionError) as excinfo:

        class MyObjectType(ObjectType):
            class Meta:
                possible_types = (dict,)

            @classmethod
            def is_type_of(cls, root, context, info):
                return False

    assert str(excinfo.value) == (
        "MyObjectType.Meta.possible_types will cause type collision with "
        "MyObjectType.is_type_of. Please use one or other."
    )


def test_objecttype_no_fields_output():
    class User(ObjectType):
        name = String()

    class Query(ObjectType):
        user = Field(User)

        def resolve_user(self, info):
            return User()

    schema = Schema(query=Query)
    result = schema.execute(
        """ query basequery {
        user {
            name
        }
    }
    """
    )
    assert not result.errors
    assert result.data == {"user": {"name": None}}


def test_abstract_objecttype_can_str():
    class MyObjectType(ObjectType):
        class Meta:
            abstract = True

        field = MyScalar()

    assert str(MyObjectType) == "MyObjectType"


def test_objecttype_meta_with_annotations():
    class Query(ObjectType):
        class Meta:
            name: str = "oops"

        hello = String()

        def resolve_hello(self, info):
            return "Hello"

    schema = Schema(query=Query)
    assert schema is not None


def test_objecttype_meta_arguments():
    class MyInterface(Interface):
        foo = String()

    class MyType(ObjectType, interfaces=[MyInterface]):
        bar = String()

    assert MyType._meta.interfaces == [MyInterface]
    assert list(MyType._meta.fields.keys()) == ["foo", "bar"]


def test_objecttype_type_name():
    class MyObjectType(ObjectType, name="FooType"):
        pass

    assert MyObjectType._meta.name == "FooType"
